@@ -3,7 +3,7 @@
  * Copyright 2022 nanocodebug <nanocodebug@gmail.com>
  * Copyright 2023 Michael Fincham <michael@hotplate.co.nz>
  * Copyright 2024 Michal Suchánek <hramrach@gmail.com>
- * Copyright 2024 Lukas F. Hartmann <lukas@mntre.com>
+ * Copyright 2024-2025 Lucie Lukas Hartmann <lukas@mntre.com>
  * Copyright 2023-2025 Johannes Schauer Marin Rodrigues <josch@mister-muffin.de>
  */
 
@@ -17,21 +17,26 @@
 #include <linux/of.h>
 #include <linux/backlight.h>
 #include <linux/version.h>
+#include <linux/reboot.h>
 
-/* abs_diff was only added to math.h in linux 6.6 */
-#ifndef abs_diff
-#define abs_diff(a, b)                                 \
-	({                                             \
-		typeof(a) __a = (a);                   \
-		typeof(b) __b = (b);                   \
-		(void)(&__a == &__b);                  \
-		__a > __b ? (__a - __b) : (__b - __a); \
-	})
-#endif
+#define MNTRE_LPC_API_UNKNOWN 0
+#define MNTRE_LPC_API_1 1
+#define MNTRE_LPC_API_2 2
+
+/* array size for lpc response buffers */
+#define LPC_RES_SZ 9
+
+typedef struct lpc_driver_data {
+	struct spi_device *spi;
+	struct power_supply *bat;
+	struct mutex lock;
+	struct backlight_device *backlight;
+	uint32_t api_version;
+} lpc_driver_data;
 
 static int lpc_probe(struct spi_device *spi);
 static void lpc_remove(struct spi_device *spi);
-static void lpc_power_off(void);
+static int lpc_power_off(struct sys_off_data *data);
 static ssize_t show_status(struct device *dev, struct device_attribute *attr,
 			   char *buf);
 static ssize_t show_cells(struct device *dev, struct device_attribute *attr,
@@ -40,35 +45,25 @@ static ssize_t show_firmware(struct device *dev, struct device_attribute *attr,
 			     char *buf);
 static ssize_t show_capacity(struct device *dev, struct device_attribute *attr,
 			     char *buf);
-
-static ssize_t lpc_command(struct device *dev, char command, uint8_t arg1,
-			   uint8_t *response);
+static ssize_t show_uart(struct device *dev, struct device_attribute *attr,
+			 char *buf);
+static ssize_t store_uart(struct device *dev, struct device_attribute *attr,
+			  const char *buf, size_t count);
+static ssize_t lpc_command(struct lpc_driver_data *lpc, char command,
+			   uint8_t arg1, uint8_t *response);
 static int get_bat_property(struct power_supply *psy,
 			    enum power_supply_property psp,
 			    union power_supply_propval *val);
 
-#define VOL_JUMP 1000000
-#define CAP_JUMP 1000000
-#define CUR_JUMP 1000000
-
-typedef struct lpc_driver_data {
-	struct spi_device *spi;
-	struct power_supply *bat;
-	struct mutex lock;
-	int last_batt_cap;
-	int last_batt_vol;
-	int last_batt_cur;
-	struct backlight_device *backlight;
-} lpc_driver_data;
-
 static DEVICE_ATTR(status, 0444, show_status, NULL);
 static DEVICE_ATTR(cells, 0444, show_cells, NULL);
 static DEVICE_ATTR(firmware, 0444, show_firmware, NULL);
 static DEVICE_ATTR(capacity, 0444, show_capacity, NULL);
+static DEVICE_ATTR(uart, 0644, show_uart, store_uart);
 
 static struct spi_board_info g_spi_board_info = {
 	.modalias = "reform2_lpc",
-	.max_speed_hz = 400000,
+	.max_speed_hz = 4000000,
 	.bus_num = 0,
 	.chip_select = 0,
 	.mode = SPI_MODE_1,
@@ -97,8 +92,6 @@ static struct power_supply_desc bat_desc = {
 
 static struct power_supply_config psy_cfg = {};
 
-static struct device *poweroff_device;
-
 static int bl_get_brightness(struct backlight_device *bl)
 {
 	u16 brightness = bl->props.brightness;
@@ -107,10 +100,9 @@ static int bl_get_brightness(struct backlight_device *bl)
 
 static int bl_update_status(struct backlight_device *bl)
 {
-	struct lpc_driver_data *data =
-		(struct lpc_driver_data *)bl_get_data(bl);
-	uint8_t buffer[8];
-	lpc_command(&data->spi->dev, 'b', bl->props.brightness, buffer);
+	struct lpc_driver_data *lpc = (struct lpc_driver_data *)bl_get_data(bl);
+	uint8_t buffer[LPC_RES_SZ];
+	lpc_command(lpc, 'b', bl->props.brightness, buffer);
 	return 0;
 }
 
@@ -136,26 +128,48 @@ static struct backlight_device *lpc_create_backlight(struct device *dev,
 
 int (*__mnt_pocket_reform_get_panel_version)(void);
 
+static uint32_t lpc_get_api_version(struct device *dev)
+{
+	int ret;
+	uint32_t version;
+	uint8_t str[LPC_RES_SZ];
+	struct lpc_driver_data *lpc =
+		(struct lpc_driver_data *)dev_get_drvdata(dev);
+
+	ret = lpc_command(lpc, 'f', 2, str);
+	if (ret)
+		return MNTRE_LPC_API_UNKNOWN;
+
+	ret = kstrtou32(str, 10, &version);
+	dev_info(dev, "version: %u (%s)\n", version, str);
+
+	if (version > 20200000 && version < 20250526) {
+		return MNTRE_LPC_API_1;
+	} else if (version >= 20250526 && version <= 30000101) {
+		return MNTRE_LPC_API_2;
+	}
+	return MNTRE_LPC_API_UNKNOWN;
+}
+
 static int lpc_probe(struct spi_device *spi)
 {
 	struct lpc_driver_data *data;
 	int ret;
 
-	printk(KERN_INFO "%s: probing ...\n", "reform2_lpc");
-
 	spi->max_speed_hz = g_spi_board_info.max_speed_hz;
 	spi->mode = g_spi_board_info.mode;
 	spi->bits_per_word = 8;
 
 	ret = spi_setup(spi);
 	if (ret) {
-		printk(KERN_ERR "%s: spi_setup failed\n", __func__);
+		dev_err(&spi->dev, "spi_setup failed\n");
 		return -ENODEV;
 	}
 
-	data = kzalloc(sizeof(struct lpc_driver_data), GFP_KERNEL);
+	data = devm_kzalloc(&spi->dev, sizeof(struct lpc_driver_data),
+			    GFP_KERNEL);
 	if (data == NULL) {
-		printk(KERN_ERR "%s: kzalloc failed\n", __func__);
+		dev_err(&spi->dev, "devm_kzalloc failed\n");
 		return -ENOMEM;
 	}
 
@@ -164,43 +178,42 @@ static int lpc_probe(struct spi_device *spi)
 	spi_set_drvdata(spi, data);
 
 	ret = device_create_file(&spi->dev, &dev_attr_status);
-	if (ret) {
-		printk(KERN_ERR "%s: device_create_file failed\n", __func__);
-	}
+	if (ret)
+		dev_err(&spi->dev, "device_create_file failed\n");
 
 	ret = device_create_file(&spi->dev, &dev_attr_cells);
-	if (ret) {
-		printk(KERN_ERR "%s: device_create_file failed\n", __func__);
-	}
+	if (ret)
+		dev_err(&spi->dev, "device_create_file failed\n");
 
 	ret = device_create_file(&spi->dev, &dev_attr_firmware);
-	if (ret) {
-		printk(KERN_ERR "%s: device_create_file failed\n", __func__);
-	}
+	if (ret)
+		dev_err(&spi->dev, "device_create_file failed\n");
 
 	ret = device_create_file(&spi->dev, &dev_attr_capacity);
-	if (ret) {
-		printk(KERN_ERR "%s: device_create_file failed\n", __func__);
-	}
+	if (ret)
+		dev_err(&spi->dev, "device_create_file failed\n");
+
+	ret = device_create_file(&spi->dev, &dev_attr_uart);
+	if (ret)
+		dev_err(&spi->dev, "device_create_file failed\n");
 
 	psy_cfg.of_node = spi->dev.of_node;
 	psy_cfg.drv_data = data;
-
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 13, 0)
-	psy_cfg.no_wakeup_source = true;
-	data->bat = power_supply_register(&spi->dev, &bat_desc, &psy_cfg);
-#else
-	data->bat = power_supply_register_no_ws(&spi->dev, &bat_desc, &psy_cfg);
-#endif
+	data->bat = devm_power_supply_register(&spi->dev, &bat_desc, &psy_cfg);
 	if (IS_ERR(data->bat)) {
-		printk(KERN_ERR "%s: power_supply_register_no_ws failed\n",
-		       __func__);
+		dev_err(&spi->dev, "dev_power_supply_register failed\n");
 		return PTR_ERR(data->bat);
 	}
 
-	// this overwrites something else that has already claimed pm_power_off on reform2 but it'll do for now
-	poweroff_device = &spi->dev;
-	pm_power_off = lpc_power_off;
+	/* register lpc as poweroff handler */
+	ret = devm_register_sys_off_handler(&spi->dev,
+					    SYS_OFF_MODE_POWER_OFF_PREPARE,
+					    SYS_OFF_PRIO_FIRMWARE,
+					    lpc_power_off, data);
+	if (ret) {
+		dev_err(&spi->dev, "devm_register_sys_off_handler failed\n");
+		return ret;
+	}
 
 	/* for MNT Pocket Reform with Display Version 2, the
 	   system controller has to control the backlight
@@ -211,188 +224,205 @@ static int lpc_probe(struct spi_device *spi)
 
 	if (__mnt_pocket_reform_get_panel_version &&
 	    __mnt_pocket_reform_get_panel_version() == 2) {
-		printk(KERN_INFO
-		       "%s: enabling backlight control for MNT Pocket Reform with Display Version 2.\n",
-		       __func__);
+		dev_info(
+			&spi->dev,
+			"enabling backlight control for MNT Pocket Reform with Display Version 2.\n");
 		data->backlight = lpc_create_backlight(&spi->dev, data);
 		if (IS_ERR(data->backlight)) {
-			printk(KERN_ERR "%s: lpc_create_backlight failed\n",
-			       __func__);
+			dev_err(&spi->dev, "lpc_create_backlight failed\n");
 		}
 	}
 
+	data->api_version = lpc_get_api_version(&spi->dev);
+
+	spi_controller_get(spi->controller);
+
 	return ret;
 }
 
 static void lpc_remove(struct spi_device *spi)
 {
-	struct lpc_driver_data *data =
-		(struct lpc_driver_data *)spi_get_drvdata(spi);
-
-	printk(KERN_INFO "%s: removing ... \n", "reform2_lpc");
-
 	device_remove_file(&spi->dev, &dev_attr_status);
 	device_remove_file(&spi->dev, &dev_attr_firmware);
 	device_remove_file(&spi->dev, &dev_attr_cells);
 	device_remove_file(&spi->dev, &dev_attr_capacity);
+	device_remove_file(&spi->dev, &dev_attr_uart);
 
-	power_supply_unregister(data->bat);
+	//spi_controller_put(spi->controller);
+}
 
-	if (pm_power_off == &lpc_power_off) {
-		pm_power_off = NULL;
+/* response[] has to have a size of at least 8 bytes! */
+static ssize_t lpc_command(struct lpc_driver_data *lpc, char command,
+			   uint8_t arg1, uint8_t *response)
+{
+	int ret = 0;
+	memset(response, 0, LPC_RES_SZ);
+
+	int delays[3] = { 50, 50, 50 };
+	if (lpc->api_version == 2) {
+		/* newer LPC firmware doesn't need huge delays */
+		/* because the response time is minimized */
+		delays[0] = 2;
+		delays[1] = 3;
+		delays[2] = 0;
 	}
 
-	kfree(data);
+	uint8_t cmd[4] = { 0xb5, command, arg1, 0x0 };
+
+	mutex_lock(&lpc->lock);
+
+	msleep(delays[0]);
+	ret = spi_write(lpc->spi, cmd, 4);
+	if (ret) {
+		dev_err(&lpc->spi->dev, "lpc_command: %c/%d spi_write failed\n",
+			command, arg1);
+		mutex_unlock(&lpc->lock);
+		return ret;
+	}
+
+	msleep(delays[1]);
+	ret = spi_read(lpc->spi, response, 8);
+	if (ret) {
+		dev_err(&lpc->spi->dev, "lpc_command: %c/%d spi_read failed\n",
+			command, arg1);
+	}
+	msleep(delays[2]);
+
+	mutex_unlock(&lpc->lock);
+	return ret;
+}
+
+static ssize_t show_uart(struct device *dev, struct device_attribute *attr,
+			 char *buf)
+{
+	/* not yet implemented */
+	return 0;
+}
+
+/* let LPC output bytes over UART to MNT Desktop Reform control panel */
+static ssize_t store_uart(struct device *dev, struct device_attribute *attr,
+			  const char *buf, size_t count)
+{
+	struct lpc_driver_data *lpc =
+		(struct lpc_driver_data *)dev_get_drvdata(dev);
+	uint8_t discard[8];
+	for (size_t i = 0; i < count; i++) {
+		lpc_command(lpc, 'z', buf[i], discard);
+	}
+	return count;
 }
 
 static ssize_t show_status(struct device *dev, struct device_attribute *attr,
 			   char *buf)
 {
+	int ret = 0;
 	uint8_t buffer[8];
 	int16_t voltage;
 	int16_t amps;
 	uint8_t percentage;
 	uint8_t status;
-	int ret = 0;
+	struct lpc_driver_data *lpc =
+		(struct lpc_driver_data *)dev_get_drvdata(dev);
 
-	ret = lpc_command(dev, 'q', 0, buffer);
-	if (ret) {
-		printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-	}
+	ret = lpc_command(lpc, 'q', 0, buffer);
+	if (ret)
+		return 0;
 
 	voltage = (int16_t)buffer[0] | ((int16_t)buffer[1] << 8);
 	amps = (int16_t)buffer[2] | ((int16_t)buffer[3] << 8);
 	percentage = buffer[4];
 	status = buffer[5];
 
-	return snprintf(buf, PAGE_SIZE, "%d.%d %d.%d %2d%% %d", voltage / 1000,
-			voltage % 1000, amps / 1000, abs(amps % 1000),
-			percentage, status);
+	return snprintf(buf, PAGE_SIZE,
+			"%d.%dV %d.%dA %2d%% [status=%d] [API=%d]\n",
+			voltage / 1000, voltage % 1000, amps / 1000,
+			abs(amps % 1000), percentage, status, lpc->api_version);
+
+	return snprintf(buf, PAGE_SIZE, "ok\n");
 }
 
 static ssize_t show_cells(struct device *dev, struct device_attribute *attr,
 			  char *buf)
 {
-	uint8_t buffer[8];
-	uint16_t cells[8];
-	ssize_t wroteChars = 0;
 	int ret = 0;
+	uint8_t buffer[LPC_RES_SZ * 2];
+	uint16_t cells[16];
+	struct lpc_driver_data *lpc =
+		(struct lpc_driver_data *)dev_get_drvdata(dev);
 
-	ret = lpc_command(dev, 'v', 0, buffer);
-	if (ret) {
-		printk(KERN_INFO "%s: lpc_command failed\n", __func__);
+	ret = lpc_command(lpc, 'v', 0, buffer);
+	if (ret)
+		return 0;
+	ret = lpc_command(lpc, 'v', 1, &buffer[8]);
+	if (ret)
+		return 0;
+
+	for (int s = 0; s < 16; s += 2) {
+		uint16_t val = buffer[s] | buffer[s + 1] << 8;
+		cells[s] = val / 1000;
+		cells[s + 1] = val % 1000;
 	}
 
-	for (uint8_t s = 0; s < 4; s++) {
-		cells[s] = buffer[s * 2] | buffer[(s * 2) + 1] << 8;
-	}
+	ret = snprintf(buf, PAGE_SIZE,
+		       "%d.%d %d.%d %d.%d %d.%d %d.%d %d.%d %d.%d %d.%d\n",
+		       cells[0], cells[1], cells[2], cells[3], cells[4],
+		       cells[5], cells[6], cells[7], cells[8], cells[9],
+		       cells[10], cells[11], cells[12], cells[13], cells[14],
+		       cells[15]);
 
-	ret = lpc_command(dev, 'v', 1, buffer);
-	if (ret) {
-		printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-	}
-
-	for (uint8_t s = 0; s < 4; s++) {
-		cells[s + 4] = buffer[s * 2] | buffer[(s * 2) + 1] << 8;
-	}
-
-	for (uint8_t s = 0; s < 8; s++) {
-		ret = snprintf(buf + wroteChars, PAGE_SIZE - wroteChars,
-			       "%d.%d ", cells[s] / 1000, cells[s] % 1000);
-		if (ret != -1) {
-			wroteChars += ret;
-		}
-	}
-
-	// drop the trailing whitespace
-	if (wroteChars > 0) {
-		wroteChars--;
-	}
-
-	return wroteChars;
+	return ret;
 }
 
 static ssize_t show_firmware(struct device *dev, struct device_attribute *attr,
 			     char *buf)
 {
-	uint8_t str1[9];
-	uint8_t str2[9];
-	uint8_t str3[9];
 	int ret = 0;
+	uint8_t str1[LPC_RES_SZ], str2[LPC_RES_SZ], str3[LPC_RES_SZ];
+	struct lpc_driver_data *lpc =
+		(struct lpc_driver_data *)dev_get_drvdata(dev);
 
-	ret = lpc_command(dev, 'f', 0, str1);
-	if (ret) {
-		printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-	}
+	ret = lpc_command(lpc, 'f', 0, str1);
+	if (ret)
+		return 0;
+	ret = lpc_command(lpc, 'f', 1, str2);
+	if (ret)
+		return 0;
+	ret = lpc_command(lpc, 'f', 2, str3);
+	if (ret)
+		return 0;
 
-	ret = lpc_command(dev, 'f', 1, str2);
-	if (ret) {
-		printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-	}
-
-	ret = lpc_command(dev, 'f', 2, str3);
-	if (ret) {
-		printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-	}
-
-	str1[8] = '\0';
-	str2[8] = '\0';
-	str3[8] = '\0';
-
-	return snprintf(buf, PAGE_SIZE, "%s %s %s", str1, str2, str3);
+	return snprintf(buf, PAGE_SIZE, "%s %s %s\n", str1, str2, str3);
 }
 
 static ssize_t show_capacity(struct device *dev, struct device_attribute *attr,
 			     char *buf)
 {
-	uint8_t buffer[8];
-	int ret = 0;
-	uint16_t cap_accu_mah, cap_min_mah, cap_max_mah;
-	ret = lpc_command(dev, 'c', 0, buffer);
-	if (ret) {
-		printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-	}
+	uint8_t buffer[LPC_RES_SZ];
+	uint16_t cap_acc_mah, cap_min_mah, cap_max_mah;
+	struct lpc_driver_data *lpc =
+		(struct lpc_driver_data *)dev_get_drvdata(dev);
 
-	cap_accu_mah = buffer[0] | (buffer[1] << 8);
+	lpc_command(lpc, 'c', 0, buffer);
+
+	cap_acc_mah = buffer[0] | (buffer[1] << 8);
 	cap_min_mah = buffer[2] | (buffer[3] << 8);
 	cap_max_mah = buffer[4] | (buffer[5] << 8);
 
-	return snprintf(buf, PAGE_SIZE, "%d %d %d", cap_accu_mah, cap_min_mah,
-			cap_max_mah);
+	return snprintf(buf, PAGE_SIZE, "[acc=%dmAh] [min=%dmAh] [max=%dmAh]\n",
+			cap_acc_mah, cap_min_mah, cap_max_mah);
 }
 
-static ssize_t lpc_command(struct device *dev, char command, uint8_t arg1,
-			   uint8_t *responseBuffer)
+static int lpc_power_off(struct sys_off_data *data)
 {
-	struct lpc_driver_data *data =
-		(struct lpc_driver_data *)dev_get_drvdata(dev);
-	uint8_t commandBuffer[4] = { 0xB5, command, arg1, 0x0 };
-	int ret = 0;
+	uint8_t buffer[LPC_RES_SZ];
+	struct lpc_driver_data *lpc = (struct lpc_driver_data *)data->cb_data;
 
-	mutex_lock(&data->lock);
-
-	ret = spi_write(data->spi, commandBuffer, 4);
-	if (ret) {
-		printk(KERN_INFO "%s: spi_write failed\n", __func__);
+	/* try to shut down power, forever */
+	while (true) {
+		lpc_command(lpc, 'p', 1, buffer);
+		msleep(100);
 	}
-	msleep(50);
-
-	ret = spi_read(data->spi, responseBuffer, 8);
-	if (ret) {
-		printk(KERN_INFO "%s: spi_read failed\n", __func__);
-	}
-	msleep(50);
-	mutex_unlock(&data->lock);
-
-	return ret;
-}
-
-static void lpc_power_off(void)
-{
-	int ret = 0;
-	uint8_t buffer[8];
-
-	ret = lpc_command(poweroff_device, 'p', 1, buffer);
+	return 0;
 }
 
 static int get_bat_property(struct power_supply *psy,
@@ -400,25 +430,23 @@ static int get_bat_property(struct power_supply *psy,
 			    union power_supply_propval *val)
 {
 	int ret = 0;
-	uint8_t buffer[8];
-	struct lpc_driver_data *data;
-	struct device *dev;
-	int16_t amp;
-
-	data = (struct lpc_driver_data *)power_supply_get_drvdata(psy);
-	dev = &(data->spi->dev);
+	uint8_t buffer[LPC_RES_SZ];
+	int milliamp, millivolt;
+	struct lpc_driver_data *lpc =
+		(struct lpc_driver_data *)power_supply_get_drvdata(psy);
 
 	switch (psp) {
 	case POWER_SUPPLY_PROP_STATUS:
 		val->intval = POWER_SUPPLY_STATUS_UNKNOWN;
-		ret = lpc_command(dev, 'q', 0, buffer);
-		if (ret) {
-			printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-		}
-		amp = (int16_t)buffer[2] | ((int16_t)buffer[3] << 8);
-		if (amp < 0) {
+		ret = lpc_command(lpc, 'q', 0, buffer);
+		if (ret)
+			return -EBUSY;
+
+		int16_t ma16 = ((int16_t)buffer[2] | ((int16_t)buffer[3] << 8));
+		milliamp = (int)ma16;
+		if (milliamp < 0) {
 			val->intval = POWER_SUPPLY_STATUS_CHARGING;
-		} else if (amp == 0) {
+		} else if (milliamp <= 100) {
 			if (buffer[4] == 100) {
 				val->intval = POWER_SUPPLY_STATUS_FULL;
 			} else {
@@ -434,99 +462,75 @@ static int get_bat_property(struct power_supply *psy,
 		break;
 
 	case POWER_SUPPLY_PROP_VOLTAGE_NOW:
-		ret = lpc_command(dev, 'q', 0, buffer);
-		if (ret) {
-			printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-			ret = -EINVAL;
-		}
-		val->intval = (buffer[0] | buffer[1] << 8) * 1000;
-		if (data->last_batt_vol &&
-		    (abs_diff(val->intval, data->last_batt_vol) > VOL_JUMP)) {
-			printk(KERN_INFO "%s: Voltage jump from %i to %i\n",
-			       __func__, data->last_batt_vol, val->intval);
-			val->intval = data->last_batt_vol +
-				      (val->intval > data->last_batt_vol ?
-					       VOL_JUMP :
-					       -VOL_JUMP);
-			printk(KERN_INFO "%s: Clamping to %i\n", __func__,
-			       val->intval);
-		}
-		data->last_batt_vol = val->intval;
+		ret = lpc_command(lpc, 'q', 0, buffer);
+		if (ret)
+			return -EBUSY;
+
+		millivolt = (buffer[0] | buffer[1] << 8);
+		if (millivolt < 5000 || millivolt >= 40000)
+			return -EBUSY;
+
+		val->intval = millivolt * 1000;
 		break;
 
 	case POWER_SUPPLY_PROP_CURRENT_NOW:
-		ret = lpc_command(dev, 'q', 0, buffer);
-		if (ret) {
-			printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-			ret = -EINVAL;
-		}
-		amp = (int16_t)buffer[2] | ((int16_t)buffer[3] << 8);
-		// negative current, battery is charging
-		// reporting a negative value is out of spec
-		if (amp < 0) {
-			amp = 0;
-		}
-		val->intval = amp * 1000;
-		if (data->last_batt_cur &&
-		    (abs_diff(val->intval, data->last_batt_cur) > CUR_JUMP)) {
-			printk(KERN_INFO "%s: Current jump from %i to %i\n",
-			       __func__, data->last_batt_cur, val->intval);
-			val->intval = data->last_batt_cur +
-				      (val->intval > data->last_batt_cur ?
-					       CUR_JUMP :
-					       -CUR_JUMP);
-			printk(KERN_INFO "%s: Clamping to %i\n", __func__,
-			       val->intval);
-		}
-		data->last_batt_cur = val->intval;
+		ret = lpc_command(lpc, 'q', 0, buffer);
+		if (ret)
+			return -EBUSY;
+
+		ma16 = (int16_t)buffer[2] | ((int16_t)buffer[3] << 8);
+		milliamp = (int)ma16;
+		if (milliamp < -20000 || milliamp >= 20000)
+			return -EBUSY;
+
+		/* negative current, battery is charging
+		   reporting a negative value is out of spec */
+		if (milliamp < 0)
+			milliamp = 0;
+		val->intval = milliamp * 1000;
+
 		break;
 
 	case POWER_SUPPLY_PROP_CAPACITY:
-		ret = lpc_command(dev, 'q', 0, buffer);
-		if (ret) {
-			printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-			ret = -EINVAL;
-		}
-		val->intval = buffer[4];
+		ret = lpc_command(lpc, 'q', 0, buffer);
+		if (ret)
+			return -EBUSY;
+
+		/* don't trigger upower emergency shutdown in case
+		 * of faulty data
+		 * (normally happens at 5% or less) */
+		int gauge = buffer[4];
+		if (gauge < 6)
+			gauge = 6;
+		if (gauge > 100)
+			return -EBUSY;
+
+		val->intval = gauge;
 		break;
 
 	case POWER_SUPPLY_PROP_CHARGE_FULL:
 	case POWER_SUPPLY_PROP_CHARGE_FULL_DESIGN:
-		ret = lpc_command(dev, 'c', 0, buffer);
-		if (ret) {
-			printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-			ret = -EINVAL;
-		}
-		val->intval = (buffer[4] | buffer[5] << 8) * 1000;
+		ret = lpc_command(lpc, 'c', 0, buffer);
+		if (ret)
+			return -EBUSY;
+
+		int milliamp_hours = (buffer[4] | buffer[5] << 8);
+		val->intval = milliamp_hours * 1000;
 		break;
 
 	case POWER_SUPPLY_PROP_CHARGE_NOW:
-		ret = lpc_command(dev, 'c', 0, buffer);
-		if (ret) {
-			printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-			ret = -EINVAL;
-		}
+		ret = lpc_command(lpc, 'c', 0, buffer);
+		if (ret)
+			return -EBUSY;
+
 		val->intval = (buffer[0] | buffer[1] << 8) * 1000;
-		if (data->last_batt_cap &&
-		    (abs_diff(val->intval, data->last_batt_cap) > CAP_JUMP)) {
-			printk(KERN_INFO "%s: Charge jump from %i to %i\n",
-			       __func__, data->last_batt_cap, val->intval);
-			val->intval = data->last_batt_cap +
-				      (val->intval > data->last_batt_cap ?
-					       CAP_JUMP :
-					       -CAP_JUMP);
-			printk(KERN_INFO "%s: Clamping to %i\n", __func__,
-			       val->intval);
-		}
-		data->last_batt_cap = val->intval;
 		break;
 
 	case POWER_SUPPLY_PROP_CHARGE_EMPTY:
-		ret = lpc_command(dev, 'c', 0, buffer);
-		if (ret) {
-			printk(KERN_INFO "%s: lpc_command failed\n", __func__);
-			ret = -EINVAL;
-		}
+		ret = lpc_command(lpc, 'c', 0, buffer);
+		if (ret)
+			return -EBUSY;
+
 		val->intval = (buffer[2] | buffer[3] << 8) * 1000;
 		break;
 
